#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__doc__ = """Adapters management package

This package provides a small set of standard base adapters for *context*, *context* and *request*, and
*context* and *request* and *view*.

See :ref:`zca` to see how PyAMS can help components management.
"""

__docformat__ = 'restructuredtext'


# import standard library
import logging
logger = logging.getLogger('PyAMS (utils)')

import venusian

# import interfaces

# import packages
from zope.interface import implementedBy


class ContextAdapter(object):
    """Context adapter"""

    def __init__(self, context):
        self.context = context


class ContextRequestAdapter(object):
    """Context + request multi-adapter"""

    def __init__(self, context, request):
        self.context = context
        self.request = request


class ContextRequestViewAdapter(object):
    """Context + request + view multi-adapter"""

    def __init__(self, context, request, view):
        self.context = context
        self.request = request
        self.view = view


class adapter_config(object):
    """Function or class decorator to declare an adapter

    Annotation parameters can be:

    :param str name: (default=''), name of the adapter
    :param [Interface...] context: an interface, or a tuple of interfaces, that the component adapts
    :param Interface provides: the interface that the adapter provides
    """

    venusian = venusian

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings.pop('for_')
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(context, name, ob):
            adapts = settings.get('context')
            if adapts is None:
                adapts = getattr(ob, '__component_adapts__', None)
                if adapts is None:
                    raise TypeError("No for argument was provided for %r and "
                                    "can't determine what the factory adapts." % ob)
            if not isinstance(adapts, tuple):
                adapts = (adapts,)

            provides = settings.get('provides')
            if provides is None:
                intfs = list(implementedBy(ob))
                if len(intfs) == 1:
                    provides = intfs[0]
                if provides is None:
                    raise TypeError("Missing 'provided' argument")

            config = context.config.with_package(info.module)
            logger.debug("Registering adapter {0} for {1} providing {2}".format(str(ob),
                                                                                str(adapts),
                                                                                str(provides)))
            config.registry.registerAdapter(ob, adapts, provides, settings.get('name', ''))

        info = self.venusian.attach(wrapped, callback, category='pyams_adapter',
                                    depth=depth + 1)

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped
